/* 
 * tplay - buffered audio player
 *
 * (c) 1997 ilkka karvinen <ik@iki.fi>
 *
 * Copyright under the GNU GENERAL PUBLIC LICENSE
 *   (see the file COPYING in this directory)
 *
 * 
 *   common functions
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include "common.h"

DWORD read_big_endian_long(byte *buffer)
{
  DWORD byte0, byte1, byte2, byte3;
  byte0 = (DWORD)buffer[0];
  byte1 = (DWORD)buffer[1];
  byte2 = (DWORD)buffer[2];
  byte3 = (DWORD)buffer[3];
  return(byte0 << 24 | byte1 << 16 | byte2 << 8 | byte3);
}

void write_big_endian_long(byte *buffer, DWORD value)
{
  buffer[0] = (byte)(value>>24 & 0xFF);
  buffer[1] = (byte)(value>>16 & 0xFF);
  buffer[2] = (byte)(value>>8  & 0xFF);
  buffer[3] = (byte)(value     & 0xFF);
}

DWORD read_little_endian_long(byte *buffer)
{
  DWORD byte0, byte1, byte2, byte3;
  byte0 = (DWORD)buffer[0];
  byte1 = (DWORD)buffer[1];
  byte2 = (DWORD)buffer[2];
  byte3 = (DWORD)buffer[3];
  return(byte3 << 24 | byte2 << 16 | byte1 << 8 | byte0);
}

WORD read_little_endian_word(byte *buffer)
{
  WORD byte0, byte1;
  byte0 = (WORD)buffer[0];
  byte1 = (WORD)buffer[1];
  return(byte1 << 8 | byte0);
}

void errprintf(char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr, fmt, ap);
}

void usage(void)
{
  printf(
 "Usage: %s [-hvDVmuxrf] [-s Hz] [-b 8|16] [-B kilobytes] [filename]\n",
	  info.progname);
  fputs("\
\n\
  -h, --help              print this help, then exit\n\
  -v, --version           print version, then exit\n\
  -V, --verbose           print useful information\n\
  -x, --swap              swap endianness\n\
  -m, --mono              mono sample\n\
  -u, --usage             print buffer usage while playing\n\
  -r, --raw               force raw audio format\n\
  -f, --force             force playing with any parameters\n\
  -s, --speed=SPEED       sampling rate (Hz)\n\
  -b, --bits=BITS         bits per sample (8 or 16)\n\
  -B, --buffer-size=SIZE  buffer size in kilobytes\n\
  -D, --device=DEVICE     device name\n", stdout);
  exit(1);
}

void warning(char *str)
{
  fprintf(stderr, "%s: Warning: %s\n", info.progname, str);
}

void die(char *str)
{
  fprintf(stderr, "%s: %s\n", info.progname, str);
  destroy_buffer();
  close_audio();
  exit(-1);
}

void errdie(char *str)
{
  fprintf(stderr, "%s: Error: %s\n", info.progname, str);
  destroy_buffer();
  close_audio();
  exit(-1);
}

void open_audio()
{
  int block_size;
  char *device;

  /* Use predefined audio device if info.device is not set     */
  /* (is NULL). Otherwise use user given audio device.         */
  if (info.device == NULL)
    device = AUDIO_DEVICE;
  else
    device = info.device;

  if ((info.audio_fd = open(device, O_WRONLY, 0)) == -1)
    errdie("Cannot open audio device");

  if (ioctl(info.audio_fd, SNDCTL_DSP_GETBLKSIZE, &block_size) < 0) 
    errdie("SNDCTL_DSP_GETBLKSIZE");

#if 0
  /* Audio device driver may give exotic audio buffer size     */
  /* for some adapters, we'd better set limits to it           */
  /* In the newer sound drivers, this should work according to */
  /* 4Front documents                                          */
  if (block_size < MIN_BLOCK_SIZE)
    block_size = MIN_BLOCK_SIZE;
  else if (block_size > MAX_BLOCK_SIZE)
    block_size = MAX_BLOCK_SIZE;
#endif

  /* Set audio block size and number of blocks */
  info.blocksize = block_size;

#ifdef DEBUG
  printf("block size: %d\n", block_size);
#endif

}

void set_audio_parameters()
{
  int format, stereo, stereo0, speed0;
  double percentage;

#ifdef DEBUG
  printf("Setting audio parameters: %d, %d, %d\n", 
	 info.speed, info.bits, info.channels);
  fflush(stdout);
#endif
  
  stereo = 1;
  switch(info.channels) {
  case 1 :
    stereo = 0;
    break;
  case 2 :
    stereo = 1;
    break;
  default :
    if (!info.force)
      errdie("Invalid number of channels");
    break;
  }

  sync_audio();

  /* Set format */
  if (info.bits == 16) {
    format = AFMT_S16_LE;
    if (ioctl(info.audio_fd, SNDCTL_DSP_SETFMT, &format) == -1)
      errdie("SNDCTL_DSP_SETFMT");

    if (!info.force && (format != AFMT_S16_LE))
      errdie("Audio device doesn't support 16 bit mode");

  } else if (info.bits == 8) { /* 8 bits requested */
    format = AFMT_U8;
    if (ioctl(info.audio_fd, SNDCTL_DSP_SETFMT, &format) == -1)
      errdie("SNDCTL_DSP_SETFMT");

    if (!info.force && (format != AFMT_U8))
      errdie("Audio device doesn't support 8 bit mode");

  }

  /* Set stereo mode */
  stereo0 = stereo;
  if (ioctl(info.audio_fd, SNDCTL_DSP_STEREO, &stereo0) == -1)
    errdie("SNDCTL_DSP_STEREO");

  if (!info.force && (stereo0 != stereo))
    errdie("Audio device doesn't support stereo mode");

  info.channels = stereo0 + 1;

  /* Set sampling speed */
  speed0 = info.speed;
  if (ioctl(info.audio_fd, SNDCTL_DSP_SPEED, &speed0) == -1)
    errdie("SNDCTL_DSP_SPEED");

  /* Q: What to do if you cannot tune audio-card on exact speed? */
  /* Author: Jerko Golubovic <jerko.golubovic@public.srce.hr> */
  /* */
  /* Variant A: you force playing even if speed is not exact */
  /*   - Don't say anything - user choosed this way with force option */
  /*     We assume he/she knows what this means..... */
  /*   - Just skip and go ahead, jump :) (100.000 Lemmings can't be wrong :) )*/
  if (!info.force) { 
  /* Variant B: you don't force playing if speed is not exact */
     /* If speed doesn't match, die and inform user about alternative */
     if (speed0 != info.speed) { 
	fprintf(stderr, "Audio device doesn't support requested speed\n");
	fprintf(stderr, " -> Requested: %d Hz, Got: %d Hz", info.speed, speed0);
	percentage = ( (speed0-info.speed)/(double)info.speed ) * 100.0;
	fprintf(stderr, " (%+dHz/%+.2f%%) \n", speed0-info.speed, percentage);
	errdie("Use --force or -f in command-line to override.");
     }
     /*	If they match, you have:       */
     /*    a) good audio-card          */
     /*    b) reasonable requirements  */
     /* Just proceed and be happy about. Don't bother user with messages :) */
  }

  info.speed = speed0;

  sync_audio();
}

void sync_audio(void)
{
  if (ioctl(info.audio_fd, SNDCTL_DSP_SYNC, NULL) < 0)
    errdie("Audio sync failed");
}

void reset_audio(void)
{
  if (ioctl(info.audio_fd, SNDCTL_DSP_RESET, 0) < 0)
    errdie("Audio reset failed");
}

void post_audio(void)
{
  if (ioctl(info.audio_fd, SNDCTL_DSP_POST, 0) < 0)
    errdie("Audio reset failed");
}

void close_audio(void)
{
  if (info.audio_fd != -1) {
    close(info.audio_fd);
    info.audio_fd = -1;
  }
}

void close_file(void)
{
  if ((info.file_fd != -1) && (info.file_fd != STDIN_FILENO)) {
    close(info.file_fd);
    info.file_fd = -1;
  }
}

void destroy_buffer(void)
{
  /* Free memory */
  if (info.buffer != NULL)
    free(info.buffer);
}

/* Write only readable characters to fp */
void nice_fputc(int c, FILE *fp)
{
  if ((c >= 32) && (c <= 127))
    putc(c, fp);
  else
    putc((int)' ', fp);
}
