/* 
 * tplay - buffered audio player  
 *
 * (c) 1997 ilkka karvinen <ik@iki.fi>
 *
 * Copyright under the GNU GENERAL PUBLIC LICENSE
 *   (see the file COPYING in this directory)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread/mit/pthread.h>
#include <getopt.h>
#include <string.h>
#include <sys/ioctl.h>
#include "common.h"


#include <linux/soundcard.h>

/* Local functions */
static void init_info(char *device);
static void allocate_buffer();
static void calculate_buffer_size(int speed, int bits, int channels);
static int  buffer_usage();
static void print_usage(int state);
static void read_header();
static void fill_buffer();
static void swap_block(byte *buffer, int blocksize);
static void producer();
static void *consumer(void *data);

/* Provided by libc */
extern char *basename(char *name);

#define OPTION_STRING "D:B:b:hms:uvVxrf"

static struct option long_options[] =
{ { "buffer-size", required_argument, NULL, 'B' },
  { "bits", required_argument, NULL, 'b' },
  { "help", no_argument, NULL, 'h' },
  { "mono", no_argument, NULL, 'm' },
  { "raw", no_argument, NULL, 'r' },
  { "speed", required_argument, NULL, 's' },
  { "usage", no_argument, NULL, 'u' },
  { "version", no_argument, NULL, 'v' },
  { "swap", no_argument, NULL, 'x' },
  { "verbose", no_argument, NULL, 'V' },
  { "force", no_argument, NULL, 'f' },
  { "device", required_argument, NULL, 'D' },
  { 0, 0, 0, 0 } };

struct info_struct info;

int verbose = FALSE;

static pthread_cond_t  cond;        /* Condition variable */
static pthread_mutex_t lock;        /* Mutex lock */

/* Initialize the buffer information */
static void init_info(char *device)
{
  info.device = device;
  info.buffer = NULL;
  info.firstblock = NULL;
  info.number_of_blocks = 0;
  info.readblock = 0;
  info.writeblock = 0;
  info.readcount = 0;
  info.writecount = 0;
  info.alldone = FALSE;
  info.overflow = FALSE;
  info.underflow = FALSE;
  info.audioset = FALSE;
  info.headerskip = 0;
  info.audio_fd = -1;
  info.file_fd = -1;
  info.blocksize = 0;
  info.bytes_on_last_block = 0;
  pthread_mutex_init(&lock, NULL);
  pthread_cond_init(&cond, NULL);
}

static void allocate_buffer()
{
  info.buffer = (byte*) malloc(info.buffer_size*sizeof(byte));
  if (info.buffer == NULL) 
    errdie("malloc failed");
  info.number_of_blocks = info.buffer_size / info.blocksize;
}

/* Buffer usage in percentage */
static int buffer_usage()
{
  int usage;
  pthread_mutex_lock(&lock);
  usage = (100 * (info.writecount - info.readcount)) / 
              info.number_of_blocks;
  pthread_mutex_unlock(&lock);
  return(usage);
}

static void calculate_buffer_size(int speed, int bits, int channels)
{
  int bytes_per_second;
  double seconds = 0.0;
  int tmp;

  bytes_per_second = speed * channels * (bits / 8);

  if (info.in_seconds) {
    tmp = (int)(1000.0 * info.seconds);
    info.buffer_size = bytes_per_second * tmp / 1000;
    seconds = info.seconds;
  } else {
    seconds = (double)info.buffer_size / (double)bytes_per_second;
  }

  if (info.buffer_size < 5*MIN_BLOCK_SIZE)
    info.buffer_size = 5*MIN_BLOCK_SIZE;

  if (verbose) 
    printf("Buffer size: %dk (%.1fs)\n", info.buffer_size / 1024, seconds);

}

static void print_usage(int state)
{
  static int prev_usage = -1;
  int usage;

  if (state != 0) {
    if ((usage = buffer_usage()) != prev_usage) {
      printf("Buffer: %d%%   \r", buffer_usage());
      fflush(stdout);
      prev_usage = usage;
    }
  } else {
    printf("Buffer: rebuf    \r");
    fflush(stdout);
  }
}

static void read_header()
{
  int bytesread, count;
  byte *p, *bufferp;

  info.firstblock = (byte*) malloc(info.blocksize*sizeof(byte));
  if (info.firstblock == NULL) 
    errdie("malloc failed");

  /* Set bufferp to point to the beginning of buffer */
  bufferp = info.firstblock;
  /* If raw format is not forced, check header */
  if (!info.forceraw) {
    /* Read one block for header */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < info.blocksize) && (count != -1) &&
	   ((count = read(info.file_fd, p, 
			  info.blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }

#ifdef DEBUG
    printf("Header: %d\n", bytesread);
#endif

    /* If less than about a Sun audio header read,     */
    /* the sample is too small.                        */
    if (bytesread < SUN_HDRSIZE)
      die("Sample size is too small");

    /* Check Sun and RIFF/WAVE audio headers */
    if (read_au(info.firstblock) && read_wav(info.firstblock)) {
      if (verbose)
	printf("Playing raw audio data: \
%d samples/s, %d bits, %d channel(s).\n", 
	       info.speed, info.bits, info.channels);
    } 

    /* Swap the block if needed. The header is swapped also but    */
    /* it is not needed anymore.                                   */
    if (info.swap)
      swap_block(bufferp, bytesread);

    /* If less bytes read than block size => all read => quit      */
    if (bytesread < info.blocksize) {
      info.alldone = TRUE;
      info.bytes_on_last_block = bytesread;
      return;
    }

    /* If header was found, read to the end of the block.          */
    if (info.headerskip) {
      count = 0;
      bytesread = info.blocksize - info.headerskip; 
      p = info.firstblock + (info.blocksize - info.headerskip);
      while ((bytesread < info.blocksize) && (count != -1) &&
	     ((count = read(info.file_fd, p, 
			    info.blocksize - bytesread)) != 0))
	{
	  p+=count; bytesread+=count;
	}
    }

    /* Advance to the next block */
    info.writeblock++;
    info.writecount++;

  } else /* Raw audio mode forced */
    if (verbose)
      printf("Playing raw audio data: \
%d samples/s, %d bits, %d channel(s).\n", 
	     info.speed, info.bits, info.channels);

  calculate_buffer_size(info.speed, info.bits, info.channels);
  allocate_buffer();

}

static void fill_buffer()
{
  int bytesread, count;
  byte *p, *bufferp;

  /* Move headerblock to buffer first */
  memmove(info.buffer, info.firstblock, info.blocksize);
  /* First block is not needed anymore so we free it */
  free(info.firstblock);

  bufferp = info.buffer + info.blocksize;
  while(TRUE) {

    /* Read a block piece by piece; especially if read from     */
    /* standard input, read may return unwanted block size.     */
    /* If count is -1, file_fd is invalid.                      */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < info.blocksize) && (count != -1) &&
	   ((count = read(info.file_fd, p, 
			  info.blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }

#ifdef DEBUG
    printf("in: block: %ld, bytesread: %d\n", 
	   info.writecount, bytesread);
#endif

    if (info.swap)
      swap_block(bufferp, bytesread);

    if (bytesread < info.blocksize) {
#ifdef DEBUG
      printf("in: last block: bytesread: %d\n", bytesread);
      fflush(stdout);
#endif
      info.bytes_on_last_block = bytesread;
      info.alldone = TRUE;
      break;
    }

    /* Advance some pointers */
    bufferp+=info.blocksize;
    info.writeblock++;
    info.writecount++;
    if (info.writeblock >= info.number_of_blocks) {
      info.writeblock = 0;
      break;
    }
  }
}

static void swap_block(byte *buffer, int blocksize)
{
  register int i;
  byte c, *p;
  p = buffer;
  for (i=0; i<blocksize/2; i++) {
    c = *p;
    *p = *(p+1);
    *++p = c;
    p++;
  }
}

static void producer()
{
  int bytesread, count;
  byte *p, *bufferp;
  
#ifdef DEBUG
  printf("Starting writer loop...\n");
  fflush(stdout);
#endif

  /* Endless loop for the audio file reader and bufferer. */
  info.overflow = FALSE;
  info.underflow = FALSE;
  bufferp = info.buffer;
  while(TRUE) {

    pthread_mutex_lock(&lock);
    if ((info.writecount - info.readcount) <= 0) {
      /* Underflow: Buffer is empty */
      info.underflow = TRUE;
#ifdef DEBUG
      printf("in: underflow\n");
      fflush(stdout);
#endif
    } else if (info.readblock == info.writeblock) {
#ifdef DEBUG
      printf("in: overflow\n");
      fflush(stdout);
#endif
      info.overflow = TRUE;
      /* Overflow: Wait for the permission to go on from consumer. */
      pthread_cond_wait(&cond, &lock);
      info.overflow = FALSE;
    }
    pthread_mutex_unlock(&lock);

    /* Read a block piece by piece; especially if read from     */
    /* standard input, read may return unwanted block size.     */
    /* If count is -1, file_fd is invalid.                      */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < info.blocksize) && (count != -1) &&
	   ((count = read(info.file_fd, p, 
			  info.blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }

    if (info.swap)
      swap_block(bufferp, bytesread);

#ifdef DEBUG2
    printf("in: %d.\n", info.writeblock);
    fflush(stdout);
#endif

    /* If less than 'blocksize' read, we assume that this is          */
    /* the last block. Save data for consumer and exit this function. */
    /* If bytesread is -1, file_fd was invalid.                       */
    if (bytesread < info.blocksize) {
#ifdef DEBUG
      printf("in: last block: bytesread: %d\n", bytesread);
      fflush(stdout);
#endif
      pthread_mutex_lock(&lock);
      info.bytes_on_last_block = bytesread;
      info.alldone = TRUE;
      pthread_mutex_unlock(&lock);
      break;
    }

    /* Advance some pointers */
    pthread_mutex_lock(&lock);
    bufferp+=info.blocksize;
    info.writeblock++;
    info.writecount++;
    if (info.writeblock >= info.number_of_blocks) {
      bufferp = info.buffer;
      info.writeblock = 0;
    }
    pthread_mutex_unlock(&lock);

  }

}

static void *consumer(void *data)
{
  byte *bufferp;
  int retry;
  int initial_sync;

#ifdef DEBUG
  printf("Starting player loop...\n");
  fflush(stdout);
#endif

  /* Endless loop for the audio player */
  bufferp = info.buffer;
  initial_sync = TRUE;
  while (TRUE) {

    if (info.show_usage)
      print_usage(1);

#ifdef DEBUG
    printf("readcount: %ld, writecount: %ld, usage: %d\n", 
	   info.readcount, info.writecount, buffer_usage());
#endif

    pthread_mutex_lock(&lock);
    if (info.alldone) {
      /* If producer is done and readcount has reached writecount, go away */
      if (info.readcount >= info.writecount) {
#ifdef DEBUG
	printf("out: last block: %ld\n", info.readcount);
	fflush(stdout);
#endif
	/* Write the last bytes and exit */
	write(info.audio_fd, bufferp, info.bytes_on_last_block);
	break;
      }
    } else if ((info.writecount - info.readcount) <= 0) {
      /* Check if we are underflowing. This usually is true here   */
      /* only when buffer is getting empty very fast.              */
      info.underflow = TRUE;
    }

    if (info.underflow) {
      /* The audio buffer is empty; give producer a break */
      info.underflow = FALSE;
      pthread_mutex_unlock(&lock);
      print_usage(0);
      sync_audio();
      for (retry=0; retry < RETRY_COUNT; retry++) {
	sleep(1);
	pthread_mutex_lock(&lock);
	if ((info.writecount - info.readcount) >= 0)
	  break;
	pthread_mutex_unlock(&lock);
      }
      if (retry == RETRY_COUNT)
	die("Retry count exceeded: Buffer empty");
    }
    pthread_mutex_unlock(&lock);

    /* Push the audio data block to device */
    write(info.audio_fd, bufferp, info.blocksize);

    /* Send an initial audio syncronisation message         */
    /* to make sure that audio driver is running with us    */
    if (initial_sync) {
      /* sync_audio(); */
      initial_sync = FALSE;
    }
#ifdef DEBUG2
    printf("out: %d.\n", info.readblock);
    fflush(stdout);
#endif

    /* Advance some pointers */
    pthread_mutex_lock(&lock);
    bufferp+=info.blocksize;
    info.readblock++;
    info.readcount++;
    if (info.readblock >= info.number_of_blocks) {
      bufferp = info.buffer;
      info.readblock = 0;
    }

    if (info.overflow) {
      /* Release the producer; it's waiting for us */
      pthread_cond_signal(&cond);
    }
    pthread_mutex_unlock(&lock);

  }

#ifdef DEBUG
  printf("out: The End.\n");
#endif

  return NULL;
}

int main(int argc, char *argv[])
{
  pthread_attr_t attr;
  pthread_t thread;
  void *retval;
  int optc;
  char *device;

  info.in_seconds = FALSE;
  info.speed    = DEFAULT_SPEED;
  info.bits     = DEFAULT_BITS;
  info.channels = DEFAULT_CHANNELS;
  info.buffer_size = BUFFER_SIZE;
  info.show_usage = FALSE;
  info.progname = basename(argv[0]);
  info.swap = FALSE;
  info.forceraw = FALSE;
  info.force = FALSE;
  device = NULL;
  while (optc = getopt_long (argc, argv, OPTION_STRING, long_options, NULL),
	 optc != EOF)
    {
      switch(optc)
	{
	case 's' :
	  info.speed = atoi(optarg);
	  break;
	case 'b' :
	  info.bits = atoi(optarg);
	  break;
	case 'm' :
	  info.channels = 1;
	  break;
	case 'B' :
#if 0
	  buffer_size = 1024*atoi(optarg);
	  /* Five times bigger buffer that blocksize should be enough */
	  if (buffer_size < 5*MIN_BLOCK_SIZE)
	    buffer_size = 5*MIN_BLOCK_SIZE;
#endif
	  if (strchr(optarg, '.')) {
	    info.in_seconds = TRUE;
	    info.seconds = atof(optarg);
	  } else {
	    info.in_seconds = FALSE;
	    info.buffer_size = 1024*atoi(optarg);
	  }
	  break;
	case 'u' :
	  info.show_usage = TRUE;
	  break;
	case 'v' :
	  printf("%s version %d.%d.%d\n", 
		 info.progname, MAJOR_VERSION, MINOR_VERSION, PATCHLEVEL);
	  exit(-1);
	  break;
	case 'V' :
	  verbose = TRUE;
	  break;
	case 'x' :
	  info.swap = TRUE;
	  break;
	case 'r' :
	  info.forceraw = TRUE;
	  break;
	case 'f' :
	  info.force = TRUE;
	  break;
	case 'D' :
	  device = (char*) malloc(64);
	  if (device != NULL)
	    strcpy(device, optarg);
	  break;
	case 'h' :
	  usage();
	  break;
	default:
	  usage();
	}
    }

#ifdef DEBUG
  printf("Inits. Buffer size: %d\n", info.buffer_size);
  fflush(stdout);
#endif

  init_info(device);

  if (optind > argc - 1) {
#ifdef DEBUG
    printf("Using stdin\n");
    fflush(stdout);
#endif
    info.file_fd = STDIN_FILENO;
  } else {
#ifdef DEBUG
    printf("File: %s\n", argv[optind]);
    fflush(stdout);
#endif
    if ((info.file_fd = open(argv[optind], O_RDONLY)) == -1) 
      errdie("Opening audio file failed");
  }

  /* Set thread attributes. Real-time and FIFO. */
  pthread_attr_init(&attr);
  pthread_attr_setschedpolicy(&attr, SCHED_FIFO);

#ifdef DEBUG
  printf("filling...");
#endif

  /* Open audio device. The device is used in setting of       */
  /* the block size and this info is needed in fill_buffer()-  */
  /* function.                                                 */
  open_audio();

  read_header();

  /* Fill the audio buffer */
  fill_buffer();

  /* Set audio parameters */
  set_audio_parameters();

#ifdef DEBUG
  printf("ready.\n");
#endif

#ifdef DEBUG
  printf("Starting the thread...\n");
  fflush(stdout);
#endif

  /* Create the thread */
  if (pthread_create(&thread, &attr, consumer, 0))
    errdie("Cannot create thread");

  /* Start producing. Don't start if sample was smaller than  */
  /* audio buffer size (alldone flag is set)                  */
  if (!info.alldone)
    producer();

  /* Close input file */
  close_file();

  /* Wait until the thread has finished. */
  pthread_join(thread, &retval);

  /* Close audio device. */
  close_audio();

  return 0;
}
  


